/*
 * Copyright (C) 2004 Jeffrey Nelson, Bernard Leach, unnamed developers
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 
 Program: iPodRead Version 0.1
 
 Description: This code returns a string typed using the buttons on the iPod.
 note: this version returns podzilla if the menu button is pressed.
 */

#include <stdlib.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <termios.h>
#include <unistd.h>
#include <errno.h>
#include <dirent.h>
#include <memory.h>

void set_keyboard(FILE *fd, int mode);
int do_completions(char *buff, int cmdMaxLength, FILE *fd, int quoteflag, char *printstr, int *pos, int posMaxi, int speed);
int do_preselected(char *buff, int cmdMaxLength, FILE *fd, char **cmdlist, int cmdmax, char *printstr, int *pos, int posMaxi, int speed);
int do_print(char *str, int pos, int posMax);
int iPodRead(char *buff, int cmdMaxLength, FILE *fd);

// TODO: get sash to pass previous command back in, so we can access it
//		 turn off echoing in sash.c so extra chars don't appear
int iPodRead(char *buff, int cmdMaxLength, FILE *fd)
{
	/*
	 Precondition: buff is a pointer to a string to return the string in.
	 cmdMaxLength is the maximum string to read.
	 fd is a pointer to the file descriptor to read from.
	 Poscondition: returns -1 if error otherwise 0
	 buff is set to string typed by user.
	 */
	pid_t childpid;
	FILE *fp;
	char input, old='\0',
		charlist[] = " abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ=+!@#$%^&*()[]{}_-<>\"\'1234567890\\/.",
		header[80],
		**cmdlist,
		*printstr;
	int place = 0,      // position of character on screen
		onechar = 0,    // 0 if no char in current place, 2 if default space, 1 otherwise
		quoteflag = 0,	// 0 if even number of quotes, 1 if odd
		currchar = 0,   // currently selected char
		speed, count,	// scroll wheel speed and counter
		cmdmax,			// the number of preselected commands
		pos, posMax;

	struct termios term;
	buff[place] = '\0';
	buff[cmdMaxLength]='\0'; // put a '\0' at the end of buff in case we fill up the rest of it without being terminated
	posMax=cmdMaxLength;
	pos=1;
	fflush(stdout);
	
	set_keyboard(fd, 0);// Turn of terminal echoing and turn off canonical mode.

	// read in our config file to get preselected commands
	if((fp=fopen("/etc/sash.conf","r"))==NULL)
	{
		set_keyboard(fd, 1);
		printf("\nCan't open /etc/sash.conf, please create it.\n");
		return -1; // abort
	}
	while(fgets(header,80,fp)!=NULL)
	{
		if(sscanf(header,"[scroll_speed]=%d",&speed)>0)
			break;
	}
	while(fgets(header,80,fp)!=NULL)
	{
		if(sscanf(header,"[command_list]=%d",&cmdmax)>0)
			break;
	}
	if((cmdlist=(char **)malloc(cmdmax*sizeof(char *)))==NULL)
	{
		fclose(fp);
		set_keyboard(fd, 1);
		return -1;
	}
	for(place=0;place<cmdmax;place++)
	{
		if((cmdlist[place]=(char *)malloc(cmdMaxLength*sizeof(char)))==NULL)
		{
			while(place--)
				free(cmdlist[place]);
			free(cmdlist);
			fclose(fp);
			set_keyboard(fd, 1);
			return -1;
		}
		if((fgets(cmdlist[place],cmdMaxLength-1,fp))==NULL) // get the preselected commands
			cmdlist[place][0]='\0';                       // note: only gets the first cmdMaxLength-1 chars of the command
		if(cmdlist[place][strlen(cmdlist[place])-1]=='\n') // strip any trailing '\n'
			cmdlist[place][strlen(cmdlist[place])-1]='\0';
	}
	fclose(fp);
	place=0;
	if((printstr=(char *)malloc(cmdMaxLength*sizeof(char)))==NULL)
	{
		set_keyboard(fd, 1);
		return -1;
	}
		
	while(1)
	{
        
        if(read(fileno(fd),&input,1) != -1)
		switch(input) {
			case 'r': // user rotated right, advance character.
				if(old!='r') {
					count=0;
					old='r';
				}
				if(++count<speed)
					break;
				(currchar==strlen(charlist)-1)?(currchar=0):(currchar++);
				if (onechar) {
					sprintf(printstr,"\b \b");
					pos=do_print(printstr,pos,posMax);
				}
				onechar = 1;
				sprintf(printstr,"%c", charlist[currchar]);
				pos=do_print(printstr,pos,posMax);
				count=0;
				break;
			case 'l': // user rotated left, decrement character.
				if(old!='l') {
					count=0;
					old='l';
				}
				if(++count<speed)
					break;
				(currchar==0)?(currchar=strlen(charlist)-1):(currchar--);
				if (onechar) {
					sprintf(printstr,"\b \b");
					pos=do_print(printstr,pos,posMax);
				}
				onechar = 1;	
				sprintf(printstr,"%c", charlist[currchar]);
				pos=do_print(printstr,pos,posMax);
				count=0;
				break;
			case 'f': // user hit fast forward, do tab-completion.
				old='f';
				if(onechar==2) {
					sprintf(printstr,"\b \b");
					pos=do_print(printstr,pos,posMax);
				}
				if(onechar == 1) {
					if((buff[place-1]=='\"')&&(!quoteflag)&&(charlist[currchar]!=' ')) {
						buff[place-1] = charlist[currchar];
						buff[place] = '\"'; // close the " again
						sprintf(printstr,"\b\b%c\"",charlist[currchar]);
						pos=do_print(printstr,pos,posMax);
					} else
						buff[place] = charlist[currchar];
					if(charlist[currchar]=='\"')
						quoteflag=!quoteflag;
					buff[++place] = '\0';
				}
				switch(do_completions(buff,cmdMaxLength,fd,quoteflag,printstr,&pos,posMax,speed)) {
					case 0: // malloc failed
						set_keyboard(fd, 1);
						printf("\n");
						fflush(stdout);
						while(cmdmax--)
							free(cmdlist[cmdmax]);
						free(cmdlist);
						return -1;
					case 1: // completion succeeded
						place=strlen(buff);
						sprintf(printstr," ");
						pos=do_print(printstr,pos,posMax);
						onechar=2;
						quoteflag=0; // successful completions always result in even numbers of quotes
						break;
					case 2: // completion aborted, return to how we were
						if(place>0) {
							if(onechar==1) {
								if(charlist[currchar]=='\"')
									quoteflag=!quoteflag;
								if(buff[--place-1]=='\"') {
									buff[place]='\0';
									buff[place-1]='\"';
									sprintf(printstr,"\b\b\"%c",charlist[currchar]);
									pos=do_print(printstr,pos,posMax);
								} else
									buff[place]='\0';
							}
							if(onechar==2) {
								sprintf(printstr," ");
								pos=do_print(printstr,pos,posMax);
							}
						}
						break;
				}
				break;
			case '\r': // user hit enter button, go to next character.
			case '\n':
				old='\r';
				if ((place==0)&&(onechar!=1)) // buffer is empty and we haven't selected anything
					break;
				if (place < cmdMaxLength-2) { // leave room for a last char and the '\0' when play is hit
					if(onechar!=1)
						buff[place]=' '; // place a ' ' by default
					else {
						if((buff[place-1]=='\"')&&(!quoteflag)&&(charlist[currchar]!=' ')) {
							buff[place-1] = charlist[currchar]; // place the char where the " was
							buff[place] = '\"'; // close the " again
							sprintf(printstr,"\b\b%c\"",charlist[currchar]);
							pos=do_print(printstr,pos,posMax);
						} else
							buff[place] = charlist[currchar]; // otherwise place the selected char
						if(charlist[currchar]=='\"')
							quoteflag=!quoteflag;
					}
					sprintf(printstr," "); // indicate that the char has been placed
					pos=do_print(printstr,pos,posMax);
					buff[++place] = '\0';
					onechar=2;
				}
				break;
			case 'w': // user hit rewind, go back a character
				old='w';
				if (onechar) {
					sprintf(printstr,"\b \b"); // erase last character
					pos=do_print(printstr,pos,posMax);
				}
				if (place) {
					currchar=strcspn(charlist,&buff[place-1]); // get the previous char index
					buff[--place]='\0'; // set the terminator back a place
					onechar=1; // look like we'd just selected that char
					if(charlist[currchar]=='\"')
						quoteflag=!quoteflag; // keep track of "s
				} else {
					buff[0]='\0'; // we're back to a null string so reset some values
					currchar=0;
					onechar=0;
				}
				break;
			case 'd': // user hit play, do the command.
				old='d';
				if (onechar == 1) { // still have last char to place
					if((buff[place-1]=='\"')&&(!quoteflag)&&(charlist[currchar]!=' ')) {
						buff[place-1] = charlist[currchar];
						buff[place] = '\"';
						sprintf(printstr,"\b\b%c\"",charlist[currchar]);
						pos=do_print(printstr,pos,posMax);
					} else
						buff[place] = charlist[currchar];
					if(charlist[currchar]=='\"')
						quoteflag=!quoteflag;
					place++;
					buff[place] = '\0';
				}
				if(quoteflag) { // close the last " if we have an odd number of them
					if(onechar==2) {
						sprintf(printstr,"\b \b");
						pos=do_print(printstr,pos,posMax);
					}
					buff[place++]='\"';
					buff[place]='\0';
					sprintf(printstr,"\"");
					pos=do_print(printstr,pos,posMax);
				}
				set_keyboard(fd, 1);
				printf("\n");
				fflush(stdout);
				while(cmdmax--)
					free(cmdlist[cmdmax]);
				free(cmdlist);
				return 0;
			case 'm': // user hit menu, go to preselected command list
				old='m';
				if(onechar==2) {
					sprintf(printstr,"\b \b");
					pos=do_print(printstr,pos,posMax);
				}
				if(onechar == 1) { // finish inputting chars
					buff[place] = charlist[currchar];
					place++;
					buff[place] = '\0';
				}
				if(do_preselected(buff,cmdMaxLength,fd,cmdlist,cmdmax,printstr,&pos,posMax,speed)) { // run the preselected command
					set_keyboard(fd, 1);
					printf("\n");
					fflush(stdout);
					while(cmdmax--)
						free(cmdlist[cmdmax]);
					free(cmdlist);
					return 0;
				} // the preselected command was aborted, was clear buff or we want to add parameters
				place=strlen(buff);
				if(place>0) {
					if(onechar==1)
						place--;
					if(onechar==2) {
						sprintf(printstr," ");
						pos=do_print(printstr,pos,posMax);
					}
				} else {
					buff[0]='\0'; // we're back to a null string so reset some values
					currchar=0;
					onechar=0;
				}
				break;
			case 'h':
				sprintf(printstr,"-- HOLD --");
				pos=do_print(printstr,pos,posMax);
				while(1)
					if(read(fileno(fd),&input,1) != -1)
						if(input=='h')
							break;
				sprintf(printstr,"\b \b");
				for(count=0;count<10;count++)
					pos=do_print(printstr,pos,posMax);
				count=0;
				break;
			default:
				set_keyboard(fd, 1);
				printf("\n");
				fflush(stdout);
				while(cmdmax--)
					free(cmdlist[cmdmax]);
				free(cmdlist);
				return -1;
		}
		fflush(stdout);
	}
} 

void set_keyboard(FILE *fd, int mode)
// fd: Input file descriptor to read from.
// mode: 
//		0: Set up keyboard input.
//		1: Load initial settings.
{
	static int calledyet = 0;
	static struct termios initial_settings, new_settings;
	// Get current modes if function has not been called yet.
	if (!calledyet) {
		tcgetattr(0, &initial_settings);
		calledyet = 1;
	}
	
	switch (mode) {
		case 0:
			tcgetattr(fileno(fd), &initial_settings);
			new_settings = initial_settings;
			new_settings.c_iflag &= ~(ICRNL | INPCK | ISTRIP | IXON); // I'm not sure how many of these
			new_settings.c_iflag |= (BRKINT);
			new_settings.c_lflag &= ~(ICANON | ECHO | ISIG | IEXTEN);		   // settings actually need setting.
			new_settings.c_oflag &= ~(OPOST);
			new_settings.c_oflag |= (ONOCR | ONLRET);
			new_settings.c_cc[VMIN] = 1;
			new_settings.c_cc[VTIME] = 0;
			tcsetattr(fileno(fd), TCSANOW, &new_settings);
			break;
		case 1:
			tcsetattr(fileno(fd), TCSAFLUSH, &initial_settings);
			break;
	}
}

int do_completions(char *buff, int cmdMaxLength, FILE *fd, int quoteflag, char *printstr, int *pos, int posMax, int speed)
{
	char input, old;
    char *head, *tail, *temp;
	char noquote[] = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ+%_-1234567890./"; // chars that don't need quoting
	char path[]="/sbin/";
    int i, j, n, m, found, stripped=0, slashed=0, count, pathflag;
	int *filter;
    struct dirent **namelist;

	if((head=(char *)malloc(cmdMaxLength*sizeof(char)))==NULL)
	{
		sprintf(printstr,"Gaaa! Not enough memory.\n");
		pos[0]=do_print(printstr,pos[0],posMax);
		return 0;
	}
	if((tail=(char *)malloc(cmdMaxLength*sizeof(char)))==NULL)
	{
		sprintf(printstr,"Gaaa! Not enough memory.\n");
		pos[0]=do_print(printstr,pos[0],posMax);
		free(head);
	  return 0;
	}
	
	if(quoteflag) // the last command in buff has no closing "
	{
		if((temp=strrchr(buff,'\"'))==NULL)
		{
			sprintf(printstr,"What the heck? Was told there was a \" but didn't find one.\n");
			pos[0]=do_print(printstr,pos[0],posMax);
			free(head);
			free(tail);
			return 2; // abort
		}
		strcpy(head,&temp[1]); // strip off leading "
		stripped++; // note how many "s we've removed
	}
	else // the last command in buff either has no "s or has both "s
	{
		if((strlen(buff)>0)&&(buff[strlen(buff)-1]=='\"'))
		{ // last command has both "s
			buff[strlen(buff)-1]='\0'; // strip the trailing "
			stripped++; // note how many "s we've removed
			if((temp=strrchr(buff,'\"'))==NULL)
			{
				sprintf(printstr,"What the heck? Was told there was a \" but didn't find one.\n");
				pos[0]=do_print(printstr,pos[0],posMax);
				free(head);
				free(tail);
				return 2; // abort
			}
			strcpy(head,&temp[1]); // strip off leading "
			stripped++; // note how many "s we've removed
			buff[strlen(buff)]='\"'; // put the trailing " back
		}
		else // last command has no "s so look for ' ' instead
		{
			if((temp=strrchr(buff,' '))==NULL)
				strcpy(head,buff); // buff only contains this command or file
			else
				strcpy(head,&temp[1]); // strip off leading ' '
		}
		// quite likely that we've just done an auto-complete to a directory and want to go down a directory
		// if not, and the diretory doesn't exist then we'll catch it with the scandir
		if((strlen(head)>0)&&(head[strlen(head)-1]!='/'))
		{
			strcpy(&head[strlen(head)],"/");
			slashed++;
		}
	}
	pathflag=0;
    if((temp=rindex(head,'/'))==NULL) // get the partially completed filename
    {
		strcpy(tail,head); // no directory is specified, so look in the path
        strcpy(head,path);
		pathflag=1;
    }
    else // separate the directory part from the incomplete filename
	{
        strcpy(tail,&temp[1]); // strip leading '/'
		head[strlen(head)-strlen(tail)]='\0'; // make head only contain the directory name
		if((n=scandir(head,&namelist,0,alphasort))<0)
		{
			// Couldn't read the specified directory, check if it was because we added a '/'
			if(tail[0]!='\0')
			{
				// It wasn't because we added a '/', some other error must have occurred so abort
				free(head);
				free(tail);
				for(i=0;i<strlen(buff);i++)
				{
					sprintf(printstr,"\b \b");
					pos[0]=do_print(printstr,pos[0],posMax);
				}
				snprintf(printstr,cmdMaxLength,"Can't read that directory!\n%s",buff);
				pos[0]=do_print(printstr,pos[0],posMax);
				return 2;
			}
			head[strlen(head)-1]='\0'; // strip trailing '/'
			slashed--;
			if((temp=rindex(head,'/'))==NULL) // get the partially completed filename
			{
				strcpy(tail,head); // no directory specified, so look in the path
				strcpy(head,path);
				pathflag=1;
			}
			else
			{
				strcpy(tail,&temp[1]); // put the partial filename into tail
				head[strlen(head)-strlen(tail)]='\0'; // shorten head
				if((n=scandir(head,&namelist,0,alphasort))<0)
				{ // scandir failed again, something must be wrong, abort
					free(head);
					free(tail);
					for(i=0;i<strlen(buff);i++)
					{
						sprintf(printstr,"\b \b");
						pos[0]=do_print(printstr,pos[0],posMax);
					}
					snprintf(printstr,cmdMaxLength,"Aaaaaaaaaaaggghh! Bugger!\n%s",buff);
					pos[0]=do_print(printstr,pos[0],posMax);
					return 2;
				}
			}
		}
	}
	if(pathflag)//strcmp(head,path)==0)
	{ // ok, need to get a listing of all completions within the path
		if((n=scandir(head,&namelist,0,alphasort))<0)
		{
			free(head);
			free(tail);
			for(i=0;i<strlen(buff);i++)
			{
				sprintf(printstr,"\b \b");
				pos[0]=do_print(printstr,pos[0],posMax);
			}
			snprintf(printstr,cmdMaxLength,"That shouldn't have failed! Does %s exist\n",path);
			pos[0]=do_print(printstr,pos[0],posMax);
			return 2;
		}
	}
	// scandir has succeeded, on with the show...
	if((filter=malloc(n*sizeof(int)))==NULL)
	{
		sprintf(printstr,"Gaaa malloc failed\n");
		pos[0]=do_print(printstr,pos[0],posMax);
		free(head);
		free(tail);
		while(n--)
			free(namelist[n]);
		free(namelist);
		return 0;
	}
	j=0;
	for(i=0;i<n;i++)
	{
		if(strncmp(tail,namelist[i]->d_name,strlen(tail))==0)
			filter[j++]=i; // filter out the non-matching directory entries
	}
	m=j;
	j=0;
	found = 0;
	if(m==0)
		found=2; // no completions exist, abort the completion
	else
	{
		for(i=strlen(tail)+stripped;i>0;i--)
		{
			sprintf(printstr,"\b \b");
			pos[0]=do_print(printstr,pos[0],posMax);
		}
		if(!pathflag)//strcmp(head,path)!=0)
		{
			for(i=strlen(head)-slashed;i>0;i--)
			{
				sprintf(printstr,"\b \b");
				pos[0]=do_print(printstr,pos[0],posMax);
			}
			snprintf(printstr,cmdMaxLength,"%s",head);
			pos[0]=do_print(printstr,pos[0],posMax);
		}
		snprintf(printstr,cmdMaxLength,"%s",namelist[filter[j]]->d_name); // display first completion without "s
		pos[0]=do_print(printstr,pos[0],posMax);
		fflush(stdout);
	}
	while(!found)
	{
		if(read(fileno(fd),&input,1) != -1)
			switch(input)
			{
				case 'r': // go to next match
					if(old!='r')
					{
						count=0;
						old='r';
					}
					if(++count<speed)
						break;
				case 'f':
					for(i=0;i<strlen(namelist[filter[j]]->d_name);i++)
					{
						sprintf(printstr,"\b \b");
						pos[0]=do_print(printstr,pos[0],posMax);
					}
					(j==m-1)?(j=0):(j++);
					snprintf(printstr,cmdMaxLength,"%s",namelist[filter[j]]->d_name);
					pos[0]=do_print(printstr,pos[0],posMax);
					fflush(stdout);
					count=0;
					break;
				case 'l': // go to previous match
					if(old!='l')
					{
						count=0;
						old='l';
					}
					if(++count<speed)
						break;
				case 'w':
					for(i=0;i<strlen(namelist[filter[j]]->d_name);i++)
					{
						sprintf(printstr,"\b \b");
						pos[0]=do_print(printstr,pos[0],posMax);
					}
					(j==0)?(j=m-1):(j--);
					snprintf(printstr,cmdMaxLength,"%s",namelist[filter[j]]->d_name);
					pos[0]=do_print(printstr,pos[0],posMax);
					fflush(stdout);
					count=0;
					break;
				case '\r':
				case '\n': // copy, at most, enough to fill up buff
					if(!pathflag)//strcmp(head,path)!=0)
						buff[strlen(buff)-strlen(head)-strlen(tail)-stripped+slashed]='\0';
					else
						buff[strlen(buff)-strlen(tail)-stripped+slashed]='\0';
					if((strspn(head,noquote)==strlen(head))&&(strspn(namelist[filter[j]]->d_name,noquote)==strlen(namelist[filter[j]]->d_name)))
						m=0; // we don't need "s
					else
						m=2; // we need "s
					if(m)
						strncpy(&buff[strlen(buff)],"\"",cmdMaxLength-strlen(buff)-1); // add beginning " if needed
					if(!pathflag)//strcmp(head,path)!=0)
						strncpy(&buff[strlen(buff)],head,cmdMaxLength-strlen(buff)-1);
					strncpy(&buff[strlen(buff)],namelist[filter[j]]->d_name,cmdMaxLength-strlen(buff)-1);
					if(m)
					{
						strncpy(&buff[strlen(buff)],"\"",cmdMaxLength-strlen(buff)-1); // add trailing " if needed
						buff[cmdMaxLength-2]='\"'; // put the " at the end if the above didn't work
						for(i=strlen(head)+strlen(namelist[filter[j]]->d_name);i>0;i--)
						{
							sprintf(printstr,"\b \b");
							pos[0]=do_print(printstr,pos[0],posMax);
						}
						snprintf(printstr,cmdMaxLength,"\"%s%s\"",head,namelist[filter[j]]->d_name);
						pos[0]=do_print(printstr,pos[0],posMax);
					}
					found=1;
					pos[0]=strlen(buff)+1;
					break;
				case 'm': // abort the completion, put back what was originally there
					for(i=0;i<strlen(namelist[filter[j]]->d_name);i++)
					{
						sprintf(printstr,"\b \b");
						pos[0]=do_print(printstr,pos[0],posMax);
					}
					if(!pathflag)//strcmp(head,path)!=0)
						for(i=strlen(head);i>0;i--)
						{
							sprintf(printstr,"\b \b");
							pos[0]=do_print(printstr,pos[0],posMax);
						}
					if(stripped)
					{
						sprintf(printstr,"\"");
						pos[0]=do_print(printstr,pos[0],posMax);
						stripped--;
					}
					if(!pathflag)//strcmp(head,path)!=0)
					{
						snprintf(printstr,cmdMaxLength,"%s",head);
						pos[0]=do_print(printstr,pos[0],posMax);
					}
					snprintf(printstr,cmdMaxLength,"%s",tail);
					pos[0]=do_print(printstr,pos[0],posMax);
					if(slashed) // remove '/' if we added one
					{
						sprintf(printstr,"\b \b");
						pos[0]=do_print(printstr,pos[0],posMax);
					}
					if(stripped)
					{
						sprintf(printstr,"\"");
						pos[0]=do_print(printstr,pos[0],posMax);
					}
					found=2;
					break;
			}
	}
	while(n--)
	{
		if(namelist[n]!=NULL)
			free(namelist[n]);
		else
			printf("free failed\n");
	}
	if(namelist!=NULL)
		free(namelist);
	if(filter!=NULL)
		free(filter);
	if(head!=NULL)
		free(head);
	if(tail!=NULL)
		free(tail);
	return found;
}


int do_preselected(char *buff, int cmdMaxLength, FILE *fd, char **cmdlist, int cmdmax, char *printstr, int *pos, int posMax, int speed)
{
	char input, old='\0';
	int cmd=0, k, count;

	if(cmdmax>0)
	{
		k=strlen(buff);
		while(k--)
		{
			sprintf(printstr,"\b \b");
			pos[0]=do_print(printstr,pos[0],posMax);
		}
		snprintf(printstr,cmdMaxLength,"%s",cmdlist[cmd]);
		pos[0]=do_print(printstr,pos[0],posMax);
		fflush(stdout);
	}
	while(1)
	{
		if(read(fileno(fd),&input,1) != -1)
		switch(input)
		{
			case 'r': // go to next preselected command
				if(old!='r')
				{
					count=0;
					old='r';
				}
				if(++count<speed)
					break;
				if(cmdmax>0)
				{
					k=strlen(cmdlist[cmd]);
					while(k--)
					{
						sprintf(printstr,"\b \b");
						pos[0]=do_print(printstr,pos[0],posMax);
					}
					(cmd==cmdmax-1)?(cmd=0):(cmd++);
					snprintf(printstr,cmdMaxLength,"%s",cmdlist[cmd]);
					pos[0]=do_print(printstr,pos[0],posMax);
					fflush(stdout);
				}
				count=0;
				break;
			case 'l': // go to previous preselected command
				if(old!='l')
				{
					count=0;
					old='l';
				}
				if(++count<speed)
					break;
				if(cmdmax>0)
				{
					k=strlen(cmdlist[cmd]);
					while(k--)
					{
						sprintf(printstr,"\b \b");
						pos[0]=do_print(printstr,pos[0],posMax);
					}
					(cmd==0)?(cmd=cmdmax-1):(cmd--);
					snprintf(printstr,cmdMaxLength,"%s",cmdlist[cmd]);
					pos[0]=do_print(printstr,pos[0],posMax);
					fflush(stdout);
				}
				count=0;
				break;
			case '\r': // centre, return but don't execute command
			case '\n':
				if(cmdmax>0)
					strncpy(buff,cmdlist[cmd],cmdMaxLength-1);
				strncpy(&buff[strlen(buff)]," ",cmdMaxLength-1);
				pos[0]=do_print(" ",pos[0],posMax);
				return 0;
			case 'w': // rew, clear the buffer
				if(cmdmax>0)
					k=strlen(cmdlist[cmd]);
				else
					k=strlen(buff);
				while(k--)
				{
					sprintf(printstr,"\b \b");
					pos[0]=do_print(printstr,pos[0],posMax);
				}
				fflush(stdout);
				buff[0]='\0';
				return 0;
			case 'f': // ffwd, not assigned yet
				break;
			case 'd': // play/pause, return and execute command
				if(cmdmax>0)
					strncpy(buff,cmdlist[cmd],cmdMaxLength-1);
				return 1;
			case 'm': // menu, abort the preselected commands
				if(cmdmax>0)
				{
					k=strlen(cmdlist[cmd]);
					while(k--)
					{
						sprintf(printstr,"\b \b");
						pos[0]=do_print(printstr,pos[0],posMax);
					}
					snprintf(printstr,cmdMaxLength,"%s",buff);
					pos[0]=do_print(printstr,pos[0],posMax);
					fflush(stdout);
				}
			return 0;
		}
	}
}

int do_print(char *str, int pos, int posMax)
{
	char *nextline="\n\033[D\033[D\033[D\033[D\033[D\033[D\033[D\033[D\033[D\033[D\033[D\033[D\033[D\033[D\033[D\033[D\033[D\033[D\033[D";
	char *prevline="\033[A\033[C\033[C\033[C\033[C\033[C\033[C\033[C\033[C\033[C\033[C\033[C\033[C\033[C\033[C\033[C\033[C\033[C\033[C\033[C";
	int i,len;
	len=strlen(str);
	for(i=0;i<len;i++)
	{
		if(str[i]=='\b')
		{
			pos--;
			if(pos<posMax-1)
				write(STDOUT_FILENO, &str[i], 1);
			if((pos%19==18)&&(pos<posMax)) // go to the end of the previous line, note reverse order to nextline
				write(STDOUT_FILENO, prevline, 60);
		}		
		else
		{
			pos++;
			if((pos%19==0)&&(pos<posMax)) // go to the start of the next line
				write(STDOUT_FILENO, nextline, 60);
			if(pos<posMax)
				write(STDOUT_FILENO, &str[i], 1);
		}
	}
	return pos;
}
